# Definition of HeAT Constants

__all__ = [
    'e',
    'Euler',
    'inf',
    'Inf',
    'Infty',
    'Infinity',
    'nan',
    'NaN',
    'pi',
]

INF = float('inf')
NAN = float('nan')
NINF = - float('inf')
PI = 3.141592653589793
E = 2.718281828459045

# aliases
inf = Inf = Infty = Infinity = INF
nan = NaN = NAN
pi = PI
e = Euler = E
